@@ -3,6 +3,8 @@ from __future__ import absolute_import
 import re
 import logging
 
+from ordereddict import OrderedDict
+
 import vertica_python.errors as errors
 
 import vertica_python.vertica.messages as messages
@@ -11,9 +13,10 @@ from vertica_python.vertica.column import Column
 logger = logging.getLogger('vertica')
 
 class Cursor(object):
-    def __init__(self, connection, cursor_type=None):
+    def __init__(self, connection, cursor_type=None, unicode_error='strict'):
         self.connection = connection
         self.cursor_type = cursor_type
+        self.unicode_error = unicode_error
         self._closed = False
         self._message = None
 
@@ -64,7 +67,7 @@ class Cursor(object):
                     # Using a regex with word boundary to correctly handle params with similar names
                     # such as :s and :start
                     match_str = u':%s\\b' % unicode(key)
-                    operation = re.sub(match_str, v.decode('utf-8'), operation, re.UNICODE)
+                    operation = re.sub(match_str, v.decode('utf-8'), operation, flags=re.UNICODE)
             elif isinstance(parameters, tuple):
                 tlist = []
                 for p in parameters:
@@ -76,7 +79,7 @@ class Cursor(object):
             else:
                 raise errors.Error("Argument 'parameters' must be dict or tuple")
 
-        self.rowcount = 0
+        self.rowcount = -1
 
         self.connection.write(messages.Query(operation))
 
@@ -88,20 +91,29 @@ class Cursor(object):
             if isinstance(message, messages.ErrorResponse):
                 raise errors.QueryError.from_error_response(message, operation)
             elif isinstance(message, messages.RowDescription):
-                self.description = map(lambda fd: Column(fd), message.fields)
-            elif (isinstance(message, messages.DataRow)
-                   or isinstance(message, messages.ReadyForQuery)):
+                self.description = map(lambda fd: Column(fd, self.unicode_error), message.fields)
+            elif isinstance(message, messages.DataRow):
+                break
+            elif isinstance(message, messages.ReadyForQuery):
                 break
             else:
                 self.connection.process_message(message)
 
     def fetchone(self):
         if isinstance(self._message, messages.DataRow):
-            self.rowcount += 1
+            if self.rowcount == -1:
+                self.rowcount = 1
+            else:
+                self.rowcount += 1
+
             row = self.row_formatter(self._message)
             # fetch next message
             self._message = self.connection.read_message()
             return row
+        elif isinstance(self._message, messages.ReadyForQuery):
+            return None
+        elif isinstance(self._message, messages.CommandComplete):
+            return None
         else:
             self.connection.process_message(self._message)
 
@@ -127,6 +139,30 @@ class Cursor(object):
     def fetchall(self):
         return list(self.iterate())
 
+    def nextset(self):
+        # skip any data for this set if exists
+        self.flush_to_command_complete()
+
+        if self._message is None:
+            return None
+        elif isinstance(self._message, messages.CommandComplete):
+            # there might be another set, read next message to find out
+            self._message = self.connection.read_message()
+            if isinstance(self._message, messages.RowDescription):
+                # next row will be either a DataRow or CommandComplete
+                self._message = self.connection.read_message()
+                return True
+            elif isinstance(self._message, messages.ReadyForQuery):
+                return None
+            else:
+                raise errors.Error('Unexpected nextset() state after CommandComplete: ' + str(self._message))
+        elif isinstance(self._message, messages.ReadyForQuery):
+            # no more sets left to be read
+            return None
+        else:
+            raise errors.Error('Unexpected nextset() state: ' + str(self._message))
+
+
     def setinputsizes(self):
         pass
 
@@ -149,36 +185,53 @@ class Cursor(object):
                 self._message = message
                 break
 
-    def copy(self, sql, data):
-        # Legacy support
-        self.copy_string(sql, data)
+    def flush_to_command_complete(self):
+        # if the last message isnt empty or CommandComplete, read messages until it is
+        if(self._message is None
+           or isinstance(self._message, messages.ReadyForQuery)
+           or isinstance(self._message, messages.CommandComplete)):
+            return
+
+        while True:
+            message = self.connection.read_message()
+            if isinstance(message, messages.CommandComplete):
+                self._message = message
+                break
+
+
+    # example:
+    #
+    # with open("/tmp/file.csv", "rb") as fs:
+    #   cursor.copy("COPY table(field1,field2) FROM STDIN DELIMITER ',' ENCLOSED BY '\"'", fs, buffer_size=65536)
+    #
+
+    def copy(self, sql, data, **kwargs):
 
-    def _copy_internal(self, sql, datagen):
         if self.closed():
             raise errors.Error('Cursor is closed')
 
+        self.flush_to_query_ready()
+
         self.connection.write(messages.Query(sql))
 
         while True:
             message = self.connection.read_message()
-            if isinstance(message, messages.ErrorResponse):
-                raise errors.QueryError.from_error_response(message, sql)
-            elif isinstance(message, messages.ReadyForQuery):
+            self.connection.process_message(message=message)
+            if isinstance(message, messages.ReadyForQuery):
                 break
             elif isinstance(message, messages.CopyInResponse):
-                # write stuff
-                for line in datagen:
-                    self.connection.write(messages.CopyData(line))
+
+                #write stuff
+                if not hasattr(data, "read"):
+                    self.connection.write(messages.CopyData(data))
+                else:
+                    # treat data as stream
+                    self.connection.write(messages.CopyStream(data, **kwargs))
+
                 self.connection.write(messages.CopyDone())
 
-    def copy_string(self, sql, data):
-        self._copy_internal(sql, [data])
-
-    def copy_file(self, sql, data, decoder=None):
-        if decoder is None:
-            self._copy_internal(sql, data)
-        else:
-            self._copy_internal(sql, (line.decode(decoder) for line in data))
+        if self.error is not None:
+            raise self.error
 
     #
     # Internal
@@ -188,16 +241,17 @@ class Cursor(object):
         return self._closed or self.connection.closed()
 
     def row_formatter(self, row_data):
-        if not self.cursor_type:
+        if self.cursor_type is None:
             return self.format_row_as_array(row_data)
-        elif self.cursor_type == 'list':
+        elif self.cursor_type in (list, 'list'):
             return self.format_row_as_array(row_data)
-        elif self.cursor_type == 'dict':
+        elif self.cursor_type in (dict, 'dict'):
             return self.format_row_as_dict(row_data)
-            # throw some error
+        else:
+            raise Exception('Unrecognized cursor_type: %r' % self.cursor_type)
 
     def format_row_as_dict(self, row_data):
-        return dict(
+        return OrderedDict(
             (self.description[idx].name, self.description[idx].convert(value))
             for idx, value in enumerate(row_data.values)
         )
